"""design module for EWOK.

Copyright (C) 2022 Theodore B. Ruegsegger
See LICENSE file for terms of use.
"""

# Indirect imports:
# from global_constants import *

from geometry import *
from kinematic_symbols import *

class SelectedItem:
    def __init__(self, figure_list, fig_index, the_figure):
        self.figure_list = figure_list
        self.fig_index = fig_index
        self.the_figure = the_figure

class S_Locus_for_3_Positions:
    def __init__(self, s_locus_center, radius_uu, sliding_hub,
                 sliding_hub_opposite, orthocenter):
        self.s_locus_center = s_locus_center
        self.radius_uu = radius_uu
        self.sliding_hub = sliding_hub
        self.sliding_hub_opposite = sliding_hub_opposite
        self.orthocenter = orthocenter

    def draw(self, canvas, pixels_per_unit, origin_cc, tag, color):
        locus = Circle_cu(self.radius_uu * pixels_per_unit)
        locus.draw(canvas, self.s_locus_center, 0, pixels_per_unit,
                   origin_cc, tag, color)
        
class LinkageDesign:

    def __init__(self, unit, pixels_per_unit, origin_cc):
        self.unit = unit
        self.pixels_per_unit = pixels_per_unit
        self.origin_cc = origin_cc
        self.centerpoints = []
        self.circlepoints = []
        self.poles = []
        self.positions = []
        self.sliderpoints = []
        self.turning_blocks = []
        self.coupler_sketches = []
        self.fixed_sketches = []
        self.coupler_border = None
        self.current_figure = None
        self.current_sketch = None
        self.m_drives_k = False
        self.mode = "Init"
        self.rotate_by_steps = False
        self.show_frame = True

    def absolute_coordinates(self, z_uc, posn_index):
        # Return fixed-plane coordinates, given
        # z_uc relative to coupler position posn_index.
        z_image = (
            z_uc * rot(self.positions[posn_index].angle)
            + self.positions[posn_index].center_uc)
        return z_image

    def closest(self, mouse_xy):
        # Finds selectable figure closest to mouse_xy
        # Selectable figures are:
        #    All coupler position specifications
        #    All centerpoints
        #    All circlepoints in design position
        # Returns (figure's list, index, figure)
        if len(self.positions) > 0:
            min_distance = abs(self.positions[0].center_uc - mouse_xy)
        elif len(self.centerpoints) > 0:
            # Centerpoint(s) can persist if all positions deleted
            min_distance = abs(self.centerpoints[0].center_uc -
                               mouse_xy)
        else:
            nearest = None
            # Just to be safe (perhaps turning blocks?)
            min_distance = float("INFINITY")
        for i in range(len(self.positions)):
            distance = abs(self.positions[i].center_uc - mouse_xy)
            if distance <= min_distance:
                min_distance = distance
                nearest = SelectedItem(
                    self.positions,
                    i,
                    self.positions[i])
        for  i in range(len(self.circlepoints)):
            distance = abs(self.absolute_coordinates(
                self.circlepoints[i].center_uc, 0)
                 - mouse_xy)
            if distance <= min_distance:
                min_distance = distance
                nearest = SelectedItem(
                    self.circlepoints,
                    i,
                    self.circlepoints[i])
        for  i in range(len(self.sliderpoints)):
            distance = abs(self.absolute_coordinates(
                self.sliderpoints[i].center_uc, 0)
                 - mouse_xy)
            if distance <= min_distance:
                min_distance = distance
                nearest = SelectedItem(
                    self.sliderpoints,
                    i,
                    self.sliderpoints[i])
        for  i in range(len(self.centerpoints)):
            distance = abs(self.centerpoints[i].center_uc - mouse_xy)
            if distance <= min_distance:
                min_distance = distance
                nearest = SelectedItem(
                    self.centerpoints,
                    i,
                    self.centerpoints[i])
        for  i in range(len(self.turning_blocks)):
            distance = abs(self.turning_blocks[i].center_uc
                           - mouse_xy)
            if distance <= min_distance:
                min_distance = distance
                nearest = SelectedItem(
                    self.turning_blocks,
                    i,
                    self.turning_blocks[i])
        return nearest
    
    def coupler_coordinates(self, z_uc, posn_index):
        # Return coordinates relative to position posn_index,
        # given z_uc in fixed-plane coordinates.
        z_image = (
            (z_uc - self.positions[posn_index].center_uc)
            * rot(-self.positions[posn_index].angle))
        return z_image

    def create_coupler_frame(self): 
        # Find limits of objects on coupler, starting with axis
        xmin_uu = ymin_uu = -10.0 / self.pixels_per_unit
        xmax_uu = 90.0 / self.pixels_per_unit
        ymax_uu = 30.0 / self.pixels_per_unit
        for figure_list in (self.circlepoints, self.sliderpoints):
            for i in range(len(figure_list)):
                xmin_uu = min(xmin_uu, figure_list[i].center_uc.real)
                xmax_uu = max(xmax_uu, figure_list[i].center_uc.real)
                ymin_uu = min(ymin_uu, figure_list[i].center_uc.imag)
                ymax_uu = max(ymax_uu, figure_list[i].center_uc.imag)
        center_uc = complex(xmax_uu + xmin_uu, ymax_uu + ymin_uu) / 2
        top_right_uu = (complex(xmax_uu - xmin_uu, ymax_uu - ymin_uu)
                        / 2)
        self.coupler_border = CouplerFrame(
            center_uc, top_right_uu, 0, self.pixels_per_unit,
            "Border", DESIGN_POSITION_COLOR)

    def create_sliderpoint(self, canvas, mouse_xy):
        fig_index = len(self.sliderpoints)
        tag = "S"+str(fig_index)
        self.mode = "MoveSlider"
        position_count = len(self.positions)
        if position_count == 1:
            s_center_uc = mouse_xy
            angle = 0
            self.sliderpoints.append(
                SliderPoint(self.coupler_coordinates(s_center_uc, 0),
                            angle, tag, DESIGN_POSITION_COLOR))
        elif position_count == 2:
            s_center_uc = mouse_xy
            angle = self.slide_direction(
                self.coupler_coordinates(s_center_uc, 0), 0, 1)
            self.sliderpoints.append(
                SliderPoint(self.coupler_coordinates(s_center_uc, 0),
                            angle, tag, DESIGN_POSITION_COLOR))
        elif position_count == 3:
            # Create the poles and the locus
            self.sliderpoint_locus = self.s_locus_3_pos()
            # Display the circle circumscribing the poles
            # self.circumpoles.draw(
            #     canvas, self.sliderpoint_locus.orthocenter, 0,
            #     self.pixels_per_unit, self.origin_cc, "PoleCircle",
            #     POLE_COLOR)
            # Display the locus
            self.sliderpoint_locus.draw(canvas, self.pixels_per_unit,
                                        self.origin_cc, "locus",
                                        DESIGN_POSITION_COLOR)
            # Mark the sliding hub:
            self.hubmarker.draw(
                canvas, self.sliderpoint_locus.sliding_hub, 0,
                self.pixels_per_unit, self.origin_cc, "locus",
                DESIGN_POSITION_COLOR)
            # Display the poles
            for i in range(len(self.poles)):
                self.poles[i].redraw(canvas, self.poles[i].center_uc,
                                     0, self.pixels_per_unit,
                                     self.origin_cc, POLE_COLOR)
            # move slider to nearest point on locus
            s_locus_center = self.sliderpoint_locus.s_locus_center
            s_center_uc = s_locus_center \
                + self.sliderpoint_locus.radius_uu * rot(
                    direction(mouse_xy - s_locus_center))
            s_center = self.coupler_coordinates(s_center_uc, 0)
            # calculate angle of sliding
            angle = self.slide_direction(s_center, 0, 1)
            self.sliderpoints.append(SliderPoint(
                s_center, angle, tag, DESIGN_POSITION_COLOR))
        self.move_slider(canvas, mouse_xy, fig_index)
        return fig_index

    def delete_crank_assembly(self, canvas, fig_index):
        # Deletes an entire crank assembly:
        #     Centerpoint
        #     Circlepoint (in all positions)
        #     Crank connecting them (in all positions)
        # Centerpoint and circlepoint must both exist
        # Erase images
        self.centerpoints[fig_index].erase(canvas)
        self.circlepoints[fig_index].erase(canvas)
        canvas.delete("Crank"+str(fig_index))
        # Delete objects from their lists
        del self.centerpoints[fig_index]
        del self.circlepoints[fig_index]
        m_count = len(self.centerpoints)
        k_count = len(self.circlepoints)
        # Reorder remaining cranks: erase, retag, redraw
        for i in range(min(m_count, k_count)):
            # Erase images
            self.centerpoints[i].erase(canvas)
            self.circlepoints[i].erase(canvas)
            canvas.delete("Crank"+str(i))
            self.centerpoints[i].tag = "M"+str(i)
            self.circlepoints[i].tag = "K"+str(i)
            self.redraw_crank_assembly(canvas, i)
        # Check for orphan M or K
        if m_count > k_count:
            last_m = m_count-1
            self.centerpoints[last_m].erase(canvas)
            self.centerpoints[last_m].tag = "M"+str(last_m)
            self.centerpoints[last_m].render(
                canvas, self.centerpoints[last_m].center_uc, 0,
                self.pixels_per_unit, self.origin_cc,
                self.centerpoints[last_m].color)
        elif k_count > m_count:
            last_k = k_count-1
            self.circlepoints[last_k].erase(canvas)
            self.circlepoints[last_k].tag = "K"+str(last_k)
            self.draw_coupler_points(
                canvas, self.circlepoints[last_k],
                self.circlepoints[last_k].color)
        # Erase any orphan crank images
        canvas.delete("Crank"+str(max(m_count, k_count)))
        # Redraw coupler frame, if enabled
        if self.show_frame:
            self.create_coupler_frame()
            self.redraw_coupler_images(
                canvas, self.coupler_border, DESIGN_POSITION_COLOR)

    def draw_coupler_images(self, canvas, the_figure, posn1_color):
        # Draw a figure's images in all positions
        # Assumes figure's center is relative to coupler
        # and that figure rotates
        for i in range(len(self.positions)):
            if i == 0:
                color = posn1_color
            else:
                color = ALT_POSITION_COLOR
            the_figure.render(
                canvas,
                self.absolute_coordinates(the_figure.center_uc, i),
                the_figure.angle + self.positions[i].angle,
                self.pixels_per_unit, self.origin_cc, color)

    def draw_coupler_points(self, canvas, the_figure, posn1_color):
        # Draw a figure's images in all positions
        # Assumes figure's center is relative to coupler
        # but that figure doesn't rotate between positions
        for i in range(len(self.positions)):
            if i == 0:
                color = posn1_color
            else:
                color = ALT_POSITION_COLOR
            the_figure.render(
                canvas,
                self.absolute_coordinates(the_figure.center_uc, i),
                the_figure.angle, self.pixels_per_unit,
                self.origin_cc, color)

    def draw_crank(self, canvas, fig_index, Kcenter_uc, color):
        # Draws the crank connecting the centerpoint
        # to a circlepoint image
        m_cc = canvas_coordinates(
            self.centerpoints[fig_index].center_uc,
            self.pixels_per_unit, self.origin_cc)
        k_cc = canvas_coordinates(Kcenter_uc, self.pixels_per_unit,
                                  self.origin_cc)
        tag = "Crank"+str(fig_index)
        canvas.create_line(
            m_cc.real, m_cc.imag,
            k_cc.real, k_cc.imag,
            tags=tag, fill=color, width=2)

    def draw_locus(self, locus, canvas, color):
        if locus.vector == (0+0j):
            return
        else:
            max_screen_diagonal_px = sqrt(canvas.winfo_width()**2 +
                                          canvas.winfo_height()**2)
            beyond_screen_uu = (max_screen_diagonal_px
                                / self.pixels_per_unit)
            ray = (
                locus.vector / abs(locus.vector)
                ) * beyond_screen_uu
            one_end_cc = canvas_coordinates(locus.origin + ray,
                                            self.pixels_per_unit,
                                            self.origin_cc)
            other_end_cc = canvas_coordinates(locus.origin - ray,
                                              self.pixels_per_unit,
                                              self.origin_cc)
        canvas.create_line(
            one_end_cc.real, one_end_cc.imag,
            other_end_cc.real, other_end_cc.imag,
            tags="locus", fill=color)
    
    def extract_design(self, canvas):
        # Return two dictionaries, one with raw, unformatted data, the
        # other with rounded, formatted data for human-readable
        # documents (e.g., html)
        def boolean_to_english(value):
            if value:
                return "Yes"
            else:
                return "No"
        e = {} # data to export
        f = {} # rounded, formatted data to export
        e['unit'] = f['unit'] = self.unit
        # filename = self.filename_from_path(self.last_file_path)
        # title = "EWOK"
        # if filename != "":
        #     title += f": {filename}"
        # f['title'] = title
        width_px = canvas.winfo_width()
        e['canvas_width'] = f['canvas_width'] = width_px
        height_px = canvas.winfo_height()
        e['canvas_height'] = f['canvas_height'] = height_px
        ppu = int(self.pixels_per_unit)
        e['pixels_per_unit'] = f['pixels_per_unit'] = ppu
        width_uu = width_px / ppu
        f['width_uu'] = width_uu
        height_uu = height_px / ppu
        f['height_uu'] = height_uu
        e['origin_cc_x'] = f['origin_cc_x'] = int(self.origin_cc.real)
        e['origin_cc_y'] = f['origin_cc_y'] = int(self.origin_cc.imag)
        e['rotate_by_steps'] = self.rotate_by_steps
        e['show_frame'] = self.show_frame
        f['rotate_by_steps'] = boolean_to_english(
            self.rotate_by_steps)
        f['show_frame'] = boolean_to_english(self.show_frame)
        e, f = self.extract_design_list(
            self.positions, 'positions', e, f)
        e, f = self.extract_design_list(
            self.centerpoints, 'centerpoints', e, f)
        e, f = self.extract_design_list(
            self.circlepoints, 'circlepoints', e, f)
        e, f = self.extract_design_list(
            self.sliderpoints, 'sliderpoints', e, f)
        e, f = self.extract_design_list(
            self.turning_blocks, 'turning_blocks', e, f)
        e, f = self.extract_design_sketches(
            self.coupler_sketches, 'coupler_sketches', e, f)
        e, f = self.extract_design_sketches(
            self.fixed_sketches, 'fixed_sketches', e, f)
        return e, f

    def extract_design_list(self, list, label, e, f):
        # Append list element details to dictionaries e and f
        list_len = len(list)
        e[label+"_count"] = f[label+"_count"] = list_len
        if list_len > 0:
            e[label] = []
            f[label] = []
        for i in range(list_len):
            ctr = list[i].center_uc
            x, y = ctr.real, ctr.imag
            xrnd, yrnd = round(x, DEFAULT_PRECISION), \
                round(y, DEFAULT_PRECISION)
            angle = list[i].angle
            anglernd = round(angle, DEFAULT_PRECISION)
            degrees = round(angle * DEGREES_PER_RADIAN,
                            DEFAULT_PRECISION)
            tag = list[i].tag
            color = list[i].color
            # Oops, "Object of type complex is not JSON serializable":
            # e[label].append({'center': ctr, 'angle': angle})
            e[label].append({'x': x, 'y': y, 'angle': angle,
                             'tag': tag, 'color': color})
            f[label].append({'x': xrnd, 'y': yrnd, 'angle': anglernd,
                             'degrees': degrees, 'tag': tag,
                             'color': color})
        return e, f

    def extract_design_sketches(self, list, label, e, f):
        # Append list element (sketch) details to dictionaries e and f
        list_len = len(list)
        e[label+"_count"] = f[label+"_count"] = list_len
        if list_len > 0:
            e[label] = []
            f[label] = []
        for i in range(list_len):
            tag = list[i].tag
            color = list[i].color
            points = list[i].geometry[0].points
            # Oops, "Object of type complex is not JSON serializable":
            # e[label].append(points)
            raw_points = []
            rounded_points = []
            for point in points:
                x, y = point.real, point.imag
                xrnd, yrnd = round(point.real, DEFAULT_PRECISION), \
                    round(point.imag, DEFAULT_PRECISION)
                raw_points.append((x,y))
                rounded_points.append((xrnd,yrnd))
            e[label].append({'tag': tag, 'color': color,
                             'points': raw_points})
            f[label].append({'tag': tag, 'color': color,
                             'points': rounded_points})
        return e, f

    def highlight_current_figure(self, canvas):
        if self.current_figure.figure_list == self.circlepoints or \
          self.current_figure.figure_list == self.sliderpoints:
            self.current_figure.the_figure.erase(canvas)
            self.draw_coupler_points(
                canvas,
                self.current_figure.the_figure,
                HIGHLIGHT_COLOR)
        else:
            self.current_figure.the_figure.redraw(
                canvas, self.current_figure.the_figure.center_uc,
                self.current_figure.the_figure.angle,
                self.pixels_per_unit, self.origin_cc, HIGHLIGHT_COLOR)

    def is_design_complete(self):
        k_count = len(self.circlepoints)
        m_count = len(self.centerpoints)
        s_count =  len(self.sliderpoints)
        return ((k_count == 2 and m_count == 2 and s_count == 0)
                or (k_count == 1 and m_count == 1 and s_count == 1)
                or (k_count == 0 and m_count == 0 and s_count == 2))
    
    def k_locus(self, m_index, posn2_index):
    # Return a circlepoint locus, given a centerpoint 
    # and an alternate coupler position.
        m_image = self.positions[0].center_uc + (
            self.centerpoints[m_index].center_uc
            - self.positions[posn2_index].center_uc
            ) * rot(self.positions[0].angle
                - self.positions[posn2_index].angle)
        return perp_bisector(
            self.centerpoints[m_index].center_uc,
            m_image)

    def m_locus(self, k_index, posn1_index, posn2_index):
    # Return the centerpoint locus, given a circlepoint 
    # and two coupler positions.
        return perp_bisector(
            self.absolute_coordinates(
                self.circlepoints[k_index].center_uc,
                posn1_index),
            self.absolute_coordinates(
                self.circlepoints[k_index].center_uc,
                posn2_index))

    def move_centerpoint(self, canvas, mouse_xy, fig_index):
        position_count = len(self.positions)
        if position_count == 0:
            self.move_centerpoint_0_pos(canvas, mouse_xy, fig_index)
        elif position_count == 1:
            self.move_centerpoint_1_pos(canvas, mouse_xy, fig_index)
        elif position_count == 2:
            self.move_centerpoint_2_pos(canvas, mouse_xy, fig_index)
        elif position_count == 3:
            self.move_centerpoint_3_pos(canvas, mouse_xy, fig_index)
        elif position_count == 4:
            self.move_centerpoint_4_pos(canvas, mouse_xy, fig_index)
        elif position_count == 5:
            self.move_centerpoint_5_pos(canvas, mouse_xy, fig_index)
    
    def move_centerpoint_0_pos(self, canvas, mouse_xy, fig_index):
        # This occurs only if last position has been deleted.
        # Centerpoint has no constraints
        self.centerpoints[fig_index].center_uc = mouse_xy
        self.current_figure = SelectedItem(
            self.centerpoints, 
            fig_index,
            self.centerpoints[fig_index])
        self.highlight_current_figure(canvas)
    
    def move_centerpoint_1_pos(self, canvas, mouse_xy, fig_index):
        self.centerpoints[fig_index].center_uc = mouse_xy
        # Centerpoint has no constraints
        # If corresponding circlepoint exists:
        if len(self.circlepoints) > fig_index:
            self.redraw_crank_assembly(canvas, fig_index)
        self.current_figure = SelectedItem(
            self.centerpoints,  fig_index,
            self.centerpoints[fig_index])
        self.highlight_current_figure(canvas)
    
    def move_centerpoint_2_pos(self, canvas, mouse_xy, fig_index):
        # If corresponding circlepoint exists:
        if len(self.circlepoints) > fig_index:
            if self.m_drives_k:
                self.centerpoints[fig_index].center_uc = mouse_xy
                # Display K locus
                circlepoint_locus = self.k_locus(fig_index, 1)
                self.redraw_locus(circlepoint_locus, canvas,
                                  K_LOCUS_COLOR)
                # Display K locus image in 2nd position
                locusImage = ComplexLine(
                    self.absolute_coordinates(
                        self.coupler_coordinates(
                            circlepoint_locus.origin, 0), 1),
                    circlepoint_locus.vector * rot(
                        self.positions[1].angle \
                        - self.positions[0].angle))
                self.draw_locus(locusImage, canvas,
                                ALT_POSITION_COLOR)
                # Move K on K locus to point closest to last K
                old_center_uc = self.circlepoints[fig_index].center_uc
                self.circlepoints[fig_index].center_uc = \
                    self.coupler_coordinates(projection(
                        self.absolute_coordinates(old_center_uc, 0),
                        circlepoint_locus), 0)
            else:
                centerpoint_locus = self. m_locus(fig_index, 0, 1)
                # Display M locus
                self.redraw_locus(centerpoint_locus, canvas,
                                  M_LOCUS_COLOR)
                # Move M on M locus to point closest to mouse_xy
                self.centerpoints[fig_index].center_uc = projection(
                    mouse_xy, centerpoint_locus)
            self.redraw_crank_assembly(canvas, fig_index)
        else:
            self.m_drives_k = True
            self.centerpoints[fig_index].center_uc = mouse_xy
            # Display K locus
            circlepoint_locus = self.k_locus(fig_index, 1)
            self.redraw_locus(circlepoint_locus, canvas,
                              K_LOCUS_COLOR)
            # Display K locus image in 2nd position
            locusImage = ComplexLine(
                self.absolute_coordinates(
                    self.coupler_coordinates(circlepoint_locus.origin,
                                             0), 1),
                circlepoint_locus.vector * rot(
                    self.positions[1].angle - self.positions[0].angle)
            )
            self.draw_locus(locusImage, canvas, ALT_POSITION_COLOR)
        self.current_figure = SelectedItem(
            self.centerpoints, 
            fig_index,
            self.centerpoints[fig_index])
        self.highlight_current_figure(canvas)
    
    def move_centerpoint_3_pos(self, canvas, mouse_xy, fig_index):
        # If no matching circlepoint exists, create it:
        if not (len(self.circlepoints) > fig_index):
            self.circlepoints.append(CirclePoint(
                mouse_xy, "K"+str(fig_index), DESIGN_POSITION_COLOR))
        self.centerpoints[fig_index].center_uc = mouse_xy
        solution = intersection(
            self.k_locus(fig_index, 1),
            self.k_locus(fig_index, 2))
        ## Need a measure of closeness
        if solution.determinant != 0:
            self.circlepoints[fig_index].center_uc = \
            self.coupler_coordinates(solution.point, 0)
        else:
            canvas.statusbar.set("%s",
                "Turning Block/Concurrency Point!")
        self.redraw_crank_assembly(canvas, fig_index)
        self.current_figure = SelectedItem(
            self.centerpoints, 
            fig_index,
            self.centerpoints[fig_index])
        self.highlight_current_figure(canvas)
    
    def move_centerpoint_4_pos(self, canvas, mouse_xy, fig_index):
        canvas.statusbar.set("%s",
            "Sorry, can't handle 4-position synthesis yet.")
        self.mode = "Select"
    
    def move_centerpoint_5_pos(self, canvas, mouse_xy, fig_index):
        canvas.statusbar.set("%s",
            "Sorry, can't handle 5-position synthesis yet.")
        self.mode = "Select"
    
    def move_circlepoint(self, canvas, mouse_xy, fig_index):
        position_count = len(self.positions)
        if position_count == 1:
            self.move_circlepoint_1_pos(canvas, mouse_xy, fig_index)
        elif position_count == 2:
            self.move_circlepoint_2_pos(canvas, mouse_xy, fig_index)
        elif position_count == 3:
            self.move_circlepoint_3_pos(canvas, mouse_xy, fig_index)
        elif position_count == 4:
            self.move_circlepoint_4_pos(canvas, mouse_xy, fig_index)
        elif position_count == 5:
            self.move_circlepoint_5_pos(canvas, mouse_xy, fig_index)
    
    def move_circlepoint_1_pos(self, canvas, mouse_xy, fig_index):
        self.circlepoints[fig_index].center_uc = \
            self.coupler_coordinates(mouse_xy, 0)
        # If corresponding centerpoint exists:
        if len(self.centerpoints) > fig_index:
            self.redraw_crank_assembly(canvas, fig_index)
        self.current_figure = SelectedItem(
            self.circlepoints, 
            fig_index,
            self.circlepoints[fig_index])
        self.highlight_current_figure(canvas)
    
    def move_circlepoint_2_pos(self, canvas, mouse_xy, fig_index):
        # If corresponding centerpoint exists:
        if len(self.centerpoints) > fig_index:
            if not self.m_drives_k:
                self.circlepoints[fig_index].center_uc = \
                    self.coupler_coordinates(mouse_xy, 0)
                # Display M locus
                centerpoint_locus = self.m_locus(fig_index, 0, 1)
                self.redraw_locus(centerpoint_locus, canvas,
                                  M_LOCUS_COLOR)
                # Move M on M locus to point closest to last M
                self.centerpoints[fig_index].center_uc = projection(
                    self.centerpoints[fig_index].center_uc,
                    centerpoint_locus)
            else:
                # Display K locus
                circlepoint_locus = self.k_locus(fig_index, 1)
                self.redraw_locus(circlepoint_locus, canvas,
                                  K_LOCUS_COLOR)
                # Display K locus image in 2nd position
                locusImage = ComplexLine(
                    self.absolute_coordinates(
                        self.coupler_coordinates(
                            circlepoint_locus.origin, 0), 1),
                    circlepoint_locus.vector * rot(
                        self.positions[1].angle \
                        - self.positions[0].angle))
                self.draw_locus(locusImage, canvas,
                                ALT_POSITION_COLOR)
                # Move K on K locus to point closest to mouse_xy
                self.circlepoints[fig_index].center_uc = \
                    self.coupler_coordinates(
                        projection(mouse_xy, circlepoint_locus), 0)
            
            self.redraw_crank_assembly(canvas, fig_index)
        else:
            self.m_drives_k = False
            self.circlepoints[fig_index].center_uc = \
                self.coupler_coordinates(mouse_xy, 0)
            # Display M locus
            self.redraw_locus(self.m_locus(fig_index, 0, 1), canvas,
                              M_LOCUS_COLOR)
            # Display K in alt positions
            # Can omit, because highlight_current_figure does it
            # self.draw_alt_images(self.circlepoints[fig_index])
        self.current_figure = SelectedItem(
            self.circlepoints, fig_index,
            self.circlepoints[fig_index])
        self.highlight_current_figure(canvas)
    
    def move_circlepoint_3_pos(self, canvas, mouse_xy, fig_index):
        # If no matching centerpoint exists, create it:
        if not (len(self.centerpoints) > fig_index):
            self.centerpoints.append(CenterPoint(
                mouse_xy, "M"+str(fig_index), FIXED_PLANE_COLOR))
        self.circlepoints[fig_index].center_uc = \
            self.coupler_coordinates(mouse_xy, 0)
        solution = intersection(self.m_locus(fig_index, 0, 1),
                                self.m_locus(fig_index, 1, 2))
        ## Need a measure of closeness
        if solution.determinant != 0:
            self.centerpoints[fig_index].center_uc = solution.point
        else:
            canvas.statusbar.set("%s", "Sliderpoint!")
        self.redraw_crank_assembly(canvas, fig_index)
        self.current_figure = SelectedItem(
            self.circlepoints, 
            fig_index,
            self.circlepoints[fig_index])
        self.highlight_current_figure(canvas)
    
    def move_circlepoint_4_pos(self, canvas, mouse_xy, fig_index):
        canvas.statusbar.set("%s",
            "Sorry, can't handle 4-position synthesis yet.")
        self.mode = "Select"
    
    def move_circlepoint_5_pos(self, canvas, mouse_xy, fig_index):
        canvas.statusbar.set("%s",
            "Sorry, can't handle 5-position synthesis yet.")
        self.mode = "Select"
    
    def move_slider(self, canvas, mouse_xy, fig_index):
        position_count = len(self.positions)
        if position_count == 1:
            self.move_slider_1_pos(canvas, mouse_xy, fig_index)
        elif position_count == 2:
            self.move_slider_2_pos(canvas, mouse_xy, fig_index)
        elif position_count == 3:
            self.move_slider_3_pos(canvas, mouse_xy, fig_index)
        elif position_count == 4:
            self.move_slider_4_pos(canvas, mouse_xy, fig_index)
        elif position_count == 5:
            self.move_slider_5_pos(canvas, mouse_xy, fig_index)
    
    def move_slider_1_pos(self, canvas, mouse_xy, fig_index):
        self.sliderpoints[fig_index].center_uc = \
            self.coupler_coordinates(mouse_xy, 0)
        self.current_figure = SelectedItem(
            self.sliderpoints,  fig_index,
            self.sliderpoints[fig_index])
        self.highlight_current_figure(canvas)
    
    def move_slider_2_pos(self, canvas, mouse_xy, fig_index):
        center_uc = self.coupler_coordinates(mouse_xy, 0)
        self.sliderpoints[fig_index].center_uc = center_uc
        self.sliderpoints[fig_index].angle = self.slide_direction(
            center_uc, 0, 1)
        # Display S in alt positions
        # Can omit, because highlight_current_figure does it
        # self.draw_alt_images(self.sliderpoints[fig_index])
        self.current_figure = SelectedItem(
            self.sliderpoints,  fig_index,
            self.sliderpoints[fig_index])
        self.highlight_current_figure(canvas)
    
    def move_slider_3_pos(self, canvas, mouse_xy, fig_index):
        # Display the locus
        self.sliderpoint_locus.draw(canvas, self.pixels_per_unit,
                                    self.origin_cc, "locus",
                                    DESIGN_POSITION_COLOR)
        # Mark the sliding hub:
        self.hubmarker.draw(canvas,
                            self.sliderpoint_locus.sliding_hub, 0,
                            self.pixels_per_unit, self.origin_cc,
                            "locus", DESIGN_POSITION_COLOR)
        # Display the poles
        for i in range(3):
            self.poles[i].redraw(canvas, self.poles[i].center_uc,
                                 0, self.pixels_per_unit,
                                 self.origin_cc, POLE_COLOR)
        # move slider to nearest point on locus
        s_locus_center = self.sliderpoint_locus.s_locus_center
        s_center_uc = s_locus_center \
            + self.sliderpoint_locus.radius_uu * rot(
                direction(mouse_xy - s_locus_center))
        s_center = self.coupler_coordinates(s_center_uc, 0)
        self.sliderpoints[fig_index].center_uc = s_center
        # calculate angle of sliding
        self.sliderpoints[fig_index].angle = self.slide_direction(
            s_center, 0, 1)
        self.current_figure = SelectedItem(
            self.sliderpoints,  fig_index,
            self.sliderpoints[fig_index])
        self.highlight_current_figure(canvas)
    
    def move_slider_4_pos(self, canvas, mouse_xy, fig_index):
        canvas.statusbar.set("%s",
            "Sorry, can't handle 4-position synthesis yet.")
        self.mode = "Select"
    
    def move_slider_5_pos(self, canvas, mouse_xy, fig_index):
        canvas.statusbar.set("%s",
            "Sorry, can't handle 5-position synthesis yet.")
        self.mode = "Select"
    
    def move_turning_block(self, canvas, mouse_xy, fig_index):
        position_count = len(self.positions)
        if position_count == 0:
            self.move_turning_block_0_pos(canvas, mouse_xy, fig_index)
        elif position_count == 1:
            self.move_turning_block_1_pos(canvas, mouse_xy, fig_index)
        elif position_count == 2:
            self.move_turning_block_2_pos(canvas, mouse_xy, fig_index)
        elif position_count == 3:
            self.move_turning_block_3_pos(canvas, mouse_xy, fig_index)
        elif position_count == 4:
            self.move_turning_block_4_pos(canvas, mouse_xy, fig_index)
        elif position_count == 5:
            self.move_turning_block_5_pos(canvas, mouse_xy, fig_index)
    
    def move_turning_block_0_pos(self, canvas, mouse_xy, fig_index):
        # This occurs only if last position has been deleted.
        # Turning block has no constraints
        self.turning_blocks[fig_index].center_uc = mouse_xy
        self.current_figure = SelectedItem(
            self.turning_blocks, fig_index,
            self.turning_blocks[fig_index])
        self.highlight_current_figure(canvas)
    
    def move_turning_block_1_pos(self, canvas, mouse_xy, fig_index):
        self.turning_blocks[fig_index].center_uc = mouse_xy
        self.current_figure = SelectedItem(
            self.turning_blocks, fig_index,
            self.turning_blocks[fig_index])
        self.highlight_current_figure(canvas)
    
    def move_turning_block_2_pos(self, canvas, mouse_xy, fig_index):
        canvas.statusbar.set("%s",
            "Beware! No calculation, just putting it where you tell me.")
        self.turning_blocks[fig_index].center_uc = mouse_xy
        self.current_figure = SelectedItem(
            self.turning_blocks, fig_index,
            self.turning_blocks[fig_index])
        self.highlight_current_figure(canvas)
    
    def move_turning_block_3_pos(self, canvas, mouse_xy, fig_index):
        canvas.statusbar.set("%s",
            "Sorry, can't handle 3-position synthesis yet.")
        self.mode = "Select"
    
    def move_turning_block_4_pos(self, canvas, mouse_xy, fig_index):
        canvas.statusbar.set("%s",
            "Sorry, can't handle 4-position synthesis yet.")
        self.mode = "Select"
    
    def move_turning_block_5_pos(self, canvas, mouse_xy, fig_index):
        canvas.statusbar.set("%s",
            "Sorry, can't handle 5-position synthesis yet.")
        self.mode = "Select"

    def pole(self, index_a, index_b):
        # Return coordinates of the pole around which the coupler
        # rotates between positions a and b, along with determinant
        center_a = self.positions[index_a].center_uc
        center_b = self.positions[index_b].center_uc
        delta_ab = center_b - center_a
        phi_ab = self.positions[index_b].angle \
            - self.positions[index_a].angle
        # pick a convenient distance for point2:
        separation = abs(delta_ab)
        point2_a = center_a + separation
        point2_b = center_b + separation * rot(phi_ab)
        perpbis1 = perp_bisector(center_a, center_b)
        perpbis2 = perp_bisector(point2_a, point2_b)
        return intersection(perpbis1, perpbis2)

    def redraw_coupler_images(self, canvas, the_figure, posn1_color):
        # Erase and redraw a figure's images in all positions
        # Assumes figure's center is relative to coupler
        # and that figure rotates
        the_figure.erase(canvas)
        self.draw_coupler_images(canvas, the_figure, posn1_color)

    def redraw_crank_assembly(self, canvas, fig_index):
        # (Re)draws an entire crank assembly:
        #     Centerpoint
        #     Circlepoint (in all positions)
        #     Crank connecting them (in all positions)
        # Centerpoint and circlepoint must both exist
        # Erase current crank images
        canvas.delete("Crank"+str(fig_index))
        # Redraw Centerpoint
        self.centerpoints[fig_index].redraw(
            canvas, self.centerpoints[fig_index].center_uc, 0,
            self.pixels_per_unit, self.origin_cc,
            self.centerpoints[fig_index].color)
        # Redraw circlepoint in all positions
        self.circlepoints[fig_index].erase(canvas)
        self.draw_coupler_points(
            canvas,
            self.circlepoints[fig_index],
            self.circlepoints[fig_index].color)
        # Display crank in all positions
        for i in range(len(self.positions)):
            if i == 0:
                color = self.circlepoints[fig_index].color
            else:
                color = ALT_POSITION_COLOR
            self.draw_crank(
                canvas, fig_index,
                self.absolute_coordinates(
                    self.circlepoints[fig_index].center_uc, i),
                color)
    
    def redraw_last_figure(self, canvas):
        if self.current_figure is None:
            pass
        elif self.current_figure.figure_list == self.circlepoints or \
          self.current_figure.figure_list == self.sliderpoints:
            self.draw_coupler_points(
                canvas, self.current_figure.the_figure,
                self.current_figure.the_figure.color)
            #######################################################
            ## Redraw crank images, if any?
            #######################################################
        else:
            self.current_figure.the_figure.redraw(
                canvas, self.current_figure.the_figure.center_uc,
                self.current_figure.the_figure.angle,
                self.pixels_per_unit, self.origin_cc,
                self.current_figure.the_figure.color)
    
    def redraw_locus(self, locus, canvas, color):
        # Erase any old locus image
        canvas.delete("locus")
        self.draw_locus(locus, canvas, color)

    def restore_from_dictionary(self, canvas, e):
        self.unit = e["unit"]
        # Resize canvas if needed
        canvas.configure(width=e["canvas_width"],
                              height=e["canvas_height"])
        canvas.update_idletasks()
        self.pixels_per_unit = e["pixels_per_unit"]
        self.origin_cc = complex(e["origin_cc_x"], e["origin_cc_y"])
        self.rotate_by_steps = e["rotate_by_steps"]
        self.show_frame = e["show_frame"]
        for i in range(e["positions_count"]):
            x, y, angle, tag, color = e["positions"][i].values()
            center_uc = complex(x, y)
            self.positions.append(Axis(center_uc, angle, tag, color))
        for i in range(e["centerpoints_count"]):
            x, y, angle, tag, color = e["centerpoints"][i].values()
            center_uc = complex(x, y)
            self.centerpoints.append(
                CenterPoint(center_uc, tag, color))
        for i in range(e["circlepoints_count"]):
            x, y, angle, tag, color = e["circlepoints"][i].values()
            center_uc = complex(x, y)
            self.circlepoints.append(
                CirclePoint(center_uc, tag, color))
        for i in range(e["sliderpoints_count"]):
            x, y, angle, tag, color = e["sliderpoints"][i].values()
            center_uc = complex(x, y)
            self.sliderpoints.append(
                SliderPoint(center_uc, angle, tag, color))
        for i in range(e["turning_blocks_count"]):
            x, y, angle, tag, color = e["turning_blocks"][i].values()
            center_uc = complex(x, y)
            self.turning_blocks.append(
                Turning_Block(center_uc, angle, tag, color))
        for i in range(e["coupler_sketches_count"]):
            tag, color, point_list = e["coupler_sketches"][i].values()
            points = []
            for pair in point_list:
                points.append(complex(pair[0], pair[1]))
            self.coupler_sketches.append(
                Sketch(0+0j, points, tag, color))
        for i in range(e["fixed_sketches_count"]):
            tag, color, point_list = e["fixed_sketches"][i].values()
            points = []
            for pair in point_list:
                points.append(complex(pair[0], pair[1]))
            self.fixed_sketches.append(
                Sketch(0+0j, points, tag, color))

    def slide_direction(self, center_uc, posn1_index, posn2_index):
        # Return angle of sliding between two positions given
        # center_uc of sliderpoint (in coupler coordinates)
        return direction(
            self.absolute_coordinates(center_uc, posn2_index)
            - self.absolute_coordinates(center_uc, posn1_index))
    
    def s_locus_2pos(self, angle, pole_uc, posn1_index, posn2_index):
        # Return a sliderpoint locus for two positions, given an angle
        # (direction of sliding)
        phi = self.positions[posn2_index].angle \
            - self.positions[posn1_index].angle
        pole = self.pole(posn1_index, posn2_index).point
        return ComplexLine(pole, rot(angle + (pi - phi)/2))

    def s_locus_3_pos(self):
        # Create the sliderpoint locus for three positions
        # Create the poles
        self.poles = []
        for i in range(3):
            center_uc = self.pole(i, (i+1)%3).point
            self.poles.append(Pole(center_uc, "Pole"+str(i),
                                     POLE_COLOR))
        # Circumscribe the pole triangle
        orthocenter = intersection(
            perp_bisector(self.poles[2].center_uc,
                          self.poles[0].center_uc),
            perp_bisector(self.poles[1].center_uc,
                          self.poles[2].center_uc)).point
        radius_uu = abs(self.poles[0].center_uc - orthocenter)
        self.circumpoles = Circle_cu(radius_uu * self.pixels_per_unit)
        s_locus_center = self.poles[0].center_uc \
            + self.poles[2].center_uc - orthocenter
        # Pick two opposite sliderpoints on the locus, determine their
        # sliding vectors, whose intersection is the sliding_hub
        point_1 = s_locus_center + radius_uu
        angle_1 = self.slide_direction(
            self.coupler_coordinates(point_1, 0), 0, 1)
        vector_1 = ComplexLine(point_1, rot(angle_1))
        point_2 = s_locus_center - radius_uu
        angle_2 = self.slide_direction(
            self.coupler_coordinates(point_2, 0), 0, 1)
        vector_2 = ComplexLine(point_2, rot(angle_2))
        sliding_hub = intersection(vector_1, vector_2).point
        sliding_hub_opposite = s_locus_center * 2 - sliding_hub
        self.hubmarker = Circle_cu(3)
        return S_Locus_for_3_Positions(
            s_locus_center, radius_uu, sliding_hub,
            sliding_hub_opposite, orthocenter)
        
    def turn_slider(self, canvas, mouse_xy, fig_index):
        position_count = len(self.positions)
        if position_count == 1:
            self.turn_slider_1_pos(canvas, mouse_xy, fig_index)
        elif position_count == 2:
            self.turn_slider_2_pos(canvas, mouse_xy, fig_index)
        elif position_count == 3:
            self.turn_slider_3_pos(canvas, mouse_xy, fig_index)
        elif position_count == 4:
            self.turn_slider_4_pos(canvas, mouse_xy, fig_index)
        elif position_count == 5:
            self.turn_slider_5_pos(canvas, mouse_xy, fig_index)
    
    def turn_slider_1_pos(self, canvas, mouse_xy, fig_index):
        # self.sliderpoints[fig_index].center_uc = \
        #     self.coupler_coordinates(mouse_xy, 0)
        self.current_figure = SelectedItem(
            self.sliderpoints, fig_index,
            self.sliderpoints[fig_index])
        self.highlight_current_figure(canvas)
    
    def turn_slider_2_pos(self, canvas, mouse_xy, fig_index):
        angle = self.sliderpoints[fig_index].angle
        pole_uc = self.pole(0, 1).point
        sliderpoint_locus = self.s_locus_2pos(angle, pole_uc, 0, 1)
        self.redraw_locus(sliderpoint_locus, canvas, S_LOCUS_COLOR)
        # Display S locus image in 2nd position
        locusImage = ComplexLine(
            self.absolute_coordinates(
                self.coupler_coordinates(
                    sliderpoint_locus.origin, 0), 1),
            sliderpoint_locus.vector * rot(
                self.positions[1].angle \
                - self.positions[0].angle))
        self.draw_locus(locusImage, canvas, ALT_POSITION_COLOR)
        # Move S to S locus preserving distance and direction from
        # pole:
        # First project old sliderpoint to S locus
        old_center = self.absolute_coordinates(
            self.sliderpoints[fig_index].center_uc, 0)
        proj_pt = projection(old_center, sliderpoint_locus)
        # Successive projections of old point on S locus will
        # gravitate toward pole (projection is always less than
        # original hypotenuse), so all we want is the direction from
        # the pole (as a unit vector)
        proj_pt_dir = rot(direction(proj_pt - pole_uc))
        # Move S to S locus preserving distance and direction from
        # pole
        distance = abs(old_center - pole_uc)
        self.sliderpoints[fig_index].center_uc = \
            self.coupler_coordinates(
                pole_uc + distance * proj_pt_dir, 0)
        self.current_figure = SelectedItem(
            self.sliderpoints, fig_index,
            self.sliderpoints[fig_index])
        self.highlight_current_figure(canvas)
    
    def turn_slider_3_pos(self, canvas, mouse_xy, fig_index):
        # Display the locus
        self.sliderpoint_locus.draw(canvas, self.pixels_per_unit,
                                    self.origin_cc, "locus",
                                    DESIGN_POSITION_COLOR)
        # Mark the sliding hub:
        self.hubmarker.draw(canvas,
                            self.sliderpoint_locus.sliding_hub, 0,
                            self.pixels_per_unit, self.origin_cc,
                            "locus", DESIGN_POSITION_COLOR)
        # Display the poles
        for i in range(3):
            self.poles[i].redraw(canvas, self.poles[i].center_uc,
                                 0, self.pixels_per_unit,
                                 self.origin_cc, POLE_COLOR)
        angle = self.sliderpoints[fig_index].angle
        sliding_vector = ComplexLine(
            self.sliderpoint_locus.sliding_hub, rot(angle))
        self.sliderpoints[fig_index].center_uc = \
            self.coupler_coordinates(projection(
            self.sliderpoint_locus.sliding_hub_opposite,
                sliding_vector), 0)
        self.current_figure = SelectedItem(
            self.sliderpoints, fig_index,
            self.sliderpoints[fig_index])
        self.highlight_current_figure(canvas)

    def turn_slider_4_pos(self, canvas, mouse_xy, fig_index):
        canvas.statusbar.set("%s",
            "Sorry, can't handle 4-position synthesis yet.")
        self.mode = "Select"
    
    def turn_slider_5_pos(self, canvas, mouse_xy, fig_index):
        canvas.statusbar.set("%s",
            "Sorry, can't handle 5-position synthesis yet.")
        self.mode = "Select"
    
    def update_all(self, canvas):
        # Redraw any fixed-frame sketches
        for sketch in self.fixed_sketches:
            sketch.redraw(canvas, 0+0j, 0, self.pixels_per_unit,
                          self.origin_cc, FIXED_PLANE_COLOR)
        for i in range(len(self.positions)):
            self.positions[i].erase(canvas)
            self.positions[i].tag = "Axis"+str(i)
            if i == 0:
                self.positions[i].color = DESIGN_POSITION_COLOR
            else:
                self.positions[i].color = ALT_POSITION_COLOR
            self.positions[i].render(
                canvas, self.positions[i].center_uc,
                self.positions[i].angle, self.pixels_per_unit,
                self.origin_cc, self.positions[i].color)
        m_count = len(self.centerpoints)
        k_count = len(self.circlepoints)
        # Erase, retag, redraw all cranks:
        for i in range(min(m_count, k_count)):
            # Erase images
            self.centerpoints[i].erase(canvas)
            self.circlepoints[i].erase(canvas)
            canvas.delete("Crank"+str(i))
            self.centerpoints[i].tag = "M"+str(i)
            self.circlepoints[i].tag = "K"+str(i)
            self.redraw_crank_assembly(canvas, i)
        if k_count > m_count:
            k = k_count - 1
            self.circlepoints[k].erase(canvas)
            self.circlepoints[k].tag = "K"+str(k)
            self.circlepoints[k].render(
                canvas, self.circlepoints[k].center_uc, 0,
                self.pixels_per_unit, self.origin_cc,
                self.circlepoints[k].color)
        elif m_count > k_count:
            m = m_count - 1
            self.centerpoints[m].erase(canvas)
            self.centerpoints[m].tag = "M"+str(m)
            self.centerpoints[m].render(
                canvas, self.centerpoints[m].center_uc, 0,
                self.pixels_per_unit, self.origin_cc,
                self.centerpoints[m].color)
        # Redraw all sliderpoints:
        for i in range(len(self.sliderpoints)):
            self.sliderpoints[i].erase(canvas)
            self.draw_coupler_points(canvas, self.sliderpoints[i],
                                     DESIGN_POSITION_COLOR) 
        # Redraw coupler frame, if enabled
        if self.show_frame and self.coupler_border is not None:
            self.coupler_border.erase(canvas)
            self.create_coupler_frame()
            self.redraw_coupler_images(
                canvas, self.coupler_border, DESIGN_POSITION_COLOR)
        # Redraw any coupler sketches
        for i in range(len(self.positions)):
            if i == 0:
                color = DESIGN_POSITION_COLOR
            else:
                color = ALT_POSITION_COLOR
            for sketch in self.coupler_sketches:
                sketch.render(
                    canvas, self.absolute_coordinates(0+0j, i),
                    self.positions[i].angle, self.pixels_per_unit,
                    self.origin_cc, color)
