"""kinematic_symbols module for EWOK.

Copyright (C) 2022 Theodore B. Ruegsegger
See LICENSE file for terms of use.
"""

from math import log10, ceil

# Indirect imports:
# from global_constants import FIXED_PLANE_COLOR

from graphics import *

class Figure:
    """Kinematic symbol, as an aggregate of shapes defined in canvas
    units but with locations in usercoordinates.

    """

    def erase(self, canvas):
        canvas.delete(self.tag)

    def redraw(self, canvas, center_uc, angle, pixels_per_unit,
               origin_cc, color):
        self.erase(canvas)
        self.render(canvas, center_uc, angle, pixels_per_unit,
                    origin_cc, color)

    def render(self, canvas, center_uc, angle, pixels_per_unit,
               origin_cc, color):
        for shape in self.geometry:
            shape.draw(canvas, center_uc, angle, pixels_per_unit,
                       origin_cc, self.tag, color)

class Axis(Figure):
     def __init__(self, center_uc, angle, tag, color):
        self.center_uc = center_uc
        self.angle = angle
        self.tag = tag
        self.color = color
        abscissa = Polyline_cu([ (-10+0j), (90+0j) ])
        ordinate = Polyline_cu([ (0+10j), (0-30j) ])
        self.geometry = [ abscissa, ordinate ]

class CenterPoint(Figure):
    def __init__(self, center_uc, tag, color):
        self.center_uc = center_uc
        self.angle = 0
        self.tag = tag
        self.color = color
        left_leg = Polyline_cu(
            [ (-11+11j), (-8+7j), (-8+0j) ])
        top = SemiCircle_cu (8)
        right_leg = Polyline_cu(
            [ (8+0j), (8+7j), (5+11j) ])
        tick2 = Polyline_cu([ (4+7j), (1+11j) ])
        tick3 = Polyline_cu([ (0+7j), (-3+11j) ])
        tick4 = Polyline_cu([ (-4+7j), (-7+11j) ])
        bearing = Circle_cu(4)
        self.geometry = [ left_leg, top, right_leg, tick2, tick3,
                          tick4, bearing ]

    def render(self, canvas, center_uc, angle, pixels_per_unit,
               origin_cc, color):
        """ Centerpoints are never rotated, angle is 0 """
        Figure.render(self, canvas, center_uc, 0, pixels_per_unit,
                      origin_cc, color)

class CirclePoint(Figure):
    def __init__(self, center_uc, tag, color):
        self.center_uc = center_uc
        self.angle = 0
        self.tag = tag
        self.color = color
        bearing = Circle_cu(3)
        bushing = Circle_cu(5)
        self.geometry = [ bearing, bushing ]

    def render(self, canvas, center_uc, angle, pixels_per_unit,
               origin_cc, color):
        """ CirclePoints are never rotated, angle is 0 """
        Figure.render(self, canvas, center_uc, 0, pixels_per_unit,
                      origin_cc, color)

class CouplerFrame(Figure):
    """Rounded rectangle to enclose coupler objects."""
    def __init__(self, center_uc, top_right_uu, angle,
                 pixels_per_unit, tag, color):
        self.center_uc = center_uc
        self.top_right_uu = top_right_uu
        self.angle = angle
        self.tag = tag
        self.color = color
        offset_cu = canvas_units(top_right_uu, pixels_per_unit)
        tr_round = QuarterCircle_cu(offset_cu, 10)
        top_edge = Polyline_cu(
            [offset_cu + (0-10j),
            -offset_cu.conjugate() + (0-10j) ] )
        tl_round = QuarterCircle_cu(-offset_cu.conjugate(), 10)
        left_edge = Polyline_cu(
            [-offset_cu.conjugate() + (-10+0j),
            -offset_cu + (-10+0j) ] )
        bl_round = QuarterCircle_cu(-offset_cu, 10)
        bottom_edge = Polyline_cu(
            [-offset_cu + (0+10j),
            offset_cu.conjugate() + (0+10j) ] )
        br_round = QuarterCircle_cu(offset_cu.conjugate(), 10)
        right_edge = Polyline_cu(
            [offset_cu.conjugate() + (10+0j),
            offset_cu + (10+0j) ] )
        self.geometry = [
            tr_round, top_edge, tl_round, left_edge,
            bl_round, bottom_edge, br_round, right_edge ]

class Dial(Figure):
    """ This is an arrow extending through a circle; not yet used """
    def __init__(self, center_uc, angle, tag, color):
        self.center_uc = center_uc
        self.angle = angle
        self.tag = tag
        self.color = color
        arrow_shaft = Polyline_cu( [ (-10+0j), (85+0j) ])
        arrow_head = Polygon_cu( [ (85+3j), (100+0j), (85-3j) ])
        bearing = Circle_cu(3)
        self.geometry = [ arrow_shaft, arrow_head, bearing ]

class Link(Figure):
    def __init__(self, center_uc, circlepoint_uc, pixels_per_unit,
                 tag, color):
        self.center_uc = center_uc
        self.angle = 0
        self.tag = tag
        self.color = color
        bar = Bar_cu(canvas_units(circlepoint_uc - center_uc,
                                  pixels_per_unit))
        self.geometry = [ bar ]

    def render(self, canvas, center_uc, angle, pixels_per_unit,
               origin_cc, color):
        ##############################################################
        # angle is zero because it's a Bar connecting two points, no
        # further rotation applies
        ##############################################################
        Figure.render(self, canvas, center_uc, 0, pixels_per_unit,
                      origin_cc, color)

class Pole(Figure):
    def __init__(self, center_uc, tag, color):
        self.center_uc = center_uc
        self.angle = 0
        self.tag = tag
        self.color = color
        inner = Circle_cu(5)
        outer = Circle_cu(7)
        self.geometry = [ inner, outer ]

    def render(self, canvas, center_uc, angle, pixels_per_unit,
               origin_cc, color):
        """ Poles are never rotated, angle is 0 """
        Figure.render(self, canvas, center_uc, 0, pixels_per_unit,
                      origin_cc, color)

class Sketch(Figure):
    """A user-drawn Polygon representing an actual shape on the
    coupler or in the fixed frame as a list of vertices in user units
    (y-axis positive downwards).

    """
    def __init__(self, center_uc, points, tag, color):
        self.center_uc = center_uc
        self.tag = tag
        self.angle = 0
        self.color = color
        shape = Polygon_uu(points)
        self.geometry = [ shape ]

class SliderPoint(Figure):
    def __init__(self, center_uc, angle, tag, color):
        self.center_uc = center_uc
        self.angle = angle
        self.tag = tag
        self.color = color
        block = Rectangle_cu ( (12-5j))
        bearing = Circle_cu(3)
        base = Polyline_cu([ (-15+12j), (-12+8j), (12+8j), (9+12j) ])
        tick2 = Polyline_cu([ (8+8j), (5+12j) ])
        tick3 = Polyline_cu([ (4+8j), (1+12j) ])
        tick4 = Polyline_cu([ (0+8j), (-3+12j) ])
        tick5 = Polyline_cu([ (-4+8j), (-7+12j) ])
        tick6 = Polyline_cu([ (-8+8j), (-11+12j) ])
        self.geometry = [ block, bearing, base,
                          tick2, tick3, tick4, tick5, tick6 ]

class TumblingBlock(Figure):
    def __init__(self, center_uc, angle, tag, color):
        self.center_uc = center_uc
        self.angle = angle
        self.tag = tag
        self.color = color
        rotor = Rectangle_cu ( (12-5j))
        centerline = Polyline_cu( [ (-20+0j), (20+0j) ])
        self.geometry = [ rotor, centerline ]

class TurningBlock(Figure):
    def __init__(self, center_uc, angle, tag, color):
        self.center_uc = center_uc
        self.angle = angle
        self.tag = tag
        self.color = color
        rotor = Rectangle_cu ( (12-5j))
        left_leg = Polyline_cu( [ (-9+14j), (-6+10j), (-6+0j) ])
        top = SemiCircle_cu (6)
        right_leg = Polyline_cu( [ (6+0j), (6+10j), (3+14j) ])
        tick2 = Polyline_cu([ (2+10j), (-1+14j) ])
        tick3 = Polyline_cu([ (-2+10j), (-5+14j) ])
        bearing = Circle_cu(3)
        self.geometry = [ rotor, left_leg, top, right_leg, tick2,
                          tick3, bearing ]

    def render(self, canvas, center_uc, angle, pixels_per_unit,
               origin_cc, color):
        # Rotate only the rotor
        for shape in self.geometry[0],:
            shape.draw(canvas, center_uc, angle, pixels_per_unit,
                       origin_cc, self.tag, color)
        # Don't rotate the remaining parts
        for shape in self.geometry[1:]:
            shape.draw(canvas, center_uc, 0, pixels_per_unit,
                       origin_cc, self.tag, color)

class FixedPlaneAxisPair():
    """Abscissa and ordinate drawn and annotated to match global
    variables (origin_cc, unit, pixels_per_unit). Spaces ticks
    sensibly based on size of maximum label."""
    
    def __init__(self, canvas, origin_cc, unit, pixels_per_unit,
                 color=FIXED_PLANE_COLOR):
        self.render(canvas, origin_cc, unit, pixels_per_unit, color)

    def erase(self, canvas):
        canvas.delete("XYAxes")

    def redraw(self, canvas, origin_cc, unit, pixels_per_unit,
               color=FIXED_PLANE_COLOR):
        self.erase(canvas)
        self.render(canvas, origin_cc, unit, pixels_per_unit, color)

    def render(self, canvas, origin_cc, unit, pixels_per_unit,
               color=FIXED_PLANE_COLOR):
        canvas.update_idletasks()
        canvas_width_px = canvas.winfo_width()
        # update_idletasks() doesn't work as advertised:
        if canvas_width_px == 1:
            canvas_width_px = canvas.winfo_reqwidth()
            canvas_height_px = canvas.winfo_reqheight()
        else:
            canvas_height_px = canvas.winfo_height()
        max_uu = max(canvas_width_px,
                     canvas_height_px) / pixels_per_unit
        tick_spacing_px = (ceil(log10(max_uu)) + len(unit) + 1) * 7
        x_org_cc = origin_cc.real
        y_org_cc = origin_cc.imag
        # abscissa through origin, whole width
        canvas.create_line(0, y_org_cc, canvas_width_px, y_org_cc,
                           width=2, fill=color,
                           arrow="last", tag="XYAxes")
        y_btm_px = y_org_cc + 10
        # annotate positive abscissa
        pos_x_uu = (canvas_width_px - x_org_cc)/pixels_per_unit
        last_x_px = x_org_cc
        for i in range(1, int(pos_x_uu) + 1):
            x_px = pixels_per_unit*i + x_org_cc
            if x_px - last_x_px > tick_spacing_px:
                canvas.create_line(x_px, y_org_cc, x_px, y_btm_px,
                                   width=2, fill=color,
                                   tag="XYAxes")
                canvas.create_text(x_px, y_btm_px + 10,
                                   text=str(i)+unit,
                                   fill=color,
                                   tag="XYAxes")
                last_x_px = x_px
        # annotate negative abscissa
        neg_x_uu = x_org_cc/pixels_per_unit
        last_x_px = x_org_cc
        for i in range(1, int(neg_x_uu) + 1):
            x_px = x_org_cc - pixels_per_unit*i
            if last_x_px - x_px > tick_spacing_px:
                canvas.create_line(x_px, y_org_cc, x_px, y_btm_px,
                                   width=2, fill=color,
                                   tag="XYAxes")
                numstr = str(-i)+unit
                canvas.create_text(x_px, y_btm_px + 10, text=numstr,
                                   fill=color,
                                   tag="XYAxes")
                last_x_px = x_px
        # ordinate through origin, whole height
        canvas.create_line(x_org_cc, canvas_height_px, x_org_cc, 0,
                           width=2, fill=color,
                           arrow="last", tag="XYAxes")
        x_lft_px = x_org_cc - 10
        # annotate positive ordinate
        pos_y_uu = y_org_cc/pixels_per_unit
        last_y_px = y_org_cc
        for i in range(1, int(pos_y_uu) + 1):
            y_px = y_org_cc - pixels_per_unit*i
            if last_y_px - y_px > tick_spacing_px:
                canvas.create_line(x_org_cc, y_px, x_lft_px, y_px,
                                   width=2, fill=color,
                                   tag="XYAxes")
                numstr = str(i)+unit
                canvas.create_text(x_lft_px-4*len(numstr), y_px,
                                   text=numstr,
                                   fill=color,
                                   tag="XYAxes")
                last_y_px = y_px
        # annotate negative ordinate
        neg_y_uu = (canvas_height_px - y_org_cc)/pixels_per_unit
        last_y_px = y_org_cc
        for i in range(1, int(neg_y_uu) + 1):
            y_px = y_org_cc + pixels_per_unit*i
            if y_px - last_y_px > tick_spacing_px:
                canvas.create_line(x_org_cc, y_px, x_lft_px, y_px,
                                   width=2, fill=color,
                                   tag="XYAxes")
                numstr = str(-i)+unit
                canvas.create_text(x_lft_px-4*len(numstr), y_px,
                                   text=numstr,
                                   fill=color,
                                   tag="XYAxes")
                last_y_px = y_px

